/*****************************************************
Firmware for raspberry pi servo board. I2C interface
to control hobby servos, and read temperature and 
IR beam states. Servo states can be read and
written over I2C bus, with a single byte of data.

Data Types:

int8_t - Signed Char 
uint16_t - Unsigned Int 
uint32_t - Unsigned Long 
int64_t - Signed Long Long

(signed/unsigned) char - 1 byte 
(signed/unsigned) short - 2 bytes 
(signed/unsigned) int - 2 bytes 
(signed/unsigned) long - 4 bytes 
(signed/unsigned) long long - 8 bytes 
float - 4 bytes (floating point) 
double - alias to float


Project : Raspberry Pi Servo Board
AVR Clock frequency: 16.00 MHz

I2C address is 0x7D, or 0x3E depending on notation
Servos located on pins PD[0-2]
IR Beam on PD3
DS18B20 Temperature sensor on PD4
LED on pin PB1

Memory address locations:
0x00 - Servo0
0x01 - Servo1
0x02 - Servo2
0x03 - IR beam
0x04 - Temperature Sensor Upper byte
0x05 - Temperature Sensor Lower byte
0x06 - Servo0 Enable/Disable

*****************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include "DS18B20.c"

#define I2C_SLAVE_ADDRESS 0x7D
#define SERVO_DATA_SIZE 7

#define STATUS_IDLE 0
#define STATUS_POINTER 1
#define STATUS_DATA 2	

#define LED_FREQ 50

#define SERVO_0_ADR 0x00
#define SERVO_ADR_END 0x02
#define IR_BEAM 0x03
#define TEMP_SENS 0x04
#define SERVO0_EN 0x06

#define IR_SENS_PIN (1<<PD3)

volatile uint8_t servoData[SERVO_DATA_SIZE];
volatile uint8_t registerPointer, receiveStatus, updateFlag, updateTemperatureFlag;

volatile uint8_t servoTimes[256];					// Waveform generation
volatile uint8_t loopflag, led_counter;

void calculateServoTimes(void);

ISR(TIMER1_COMPA_vect) {
	// Allow servo timing sequence to start
	loopflag = 0;
	led_counter++;
	
	// Blink LED
	if (led_counter > LED_FREQ) {
		PORTB ^= (1<<PB1);
		led_counter = 0;
	}
}

// TWI interrupt handler
ISR(TWI_vect) {
	uint8_t status;
	/* Check which TWI event has happened. Works as slave transmitter or receiver. Data byte can be
	written to indicate which register to read from or write to.
	*/
	
	// Read the status register
	status = (TWSR & 0xF8);
	
	switch(status) {
	
		/*************************************** Slave Receive mode*************************************/
		// Own SLA+W has been received; ACK has been returned
		case 0x60 : {
			// Get ready for pointer byte
			receiveStatus = STATUS_POINTER;
			// Wait for ACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		// Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
		case 0x68 : {
			// Get ready for pointer byte
			receiveStatus = STATUS_POINTER;
			// Wait for ACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		// Previously addressed with own SLA+W; data has been received; ACK has been returned
		case 0x80 : {
			if (receiveStatus == STATUS_POINTER) {
				// First byte of transmission, register pointer
				registerPointer = TWDR;
				if (registerPointer >= SERVO_DATA_SIZE) {
					registerPointer = 0;
				}
				receiveStatus = STATUS_DATA;
				// Wait for ACK, clear interrupt flag
				TWCR |= (1<<TWINT)|(1<<TWEA);
				
			}else if (receiveStatus == STATUS_DATA) {
				// Check if address received is for servo setting, or for performing a sensor refresh
				if (registerPointer <= SERVO_ADR_END) {
					// Second byte of transmission, servo data
					servoData[registerPointer] = TWDR;
					updateFlag = 1;
				
				}else {
					// Either temperature or IR beam requested
					if (registerPointer == TEMP_SENS) {
						// Request a temperature reading update
						updateTemperatureFlag = 1;
					}else if (registerPointer == IR_BEAM) {
						// Read pin state for IR beam sensor
						servoData[registerPointer] = (PIND & IR_SENS_PIN);
					}else if (registerPointer == SERVO0_EN) {
						// Enable or disable Servo0 output
						servoData[registerPointer] = TWDR;
					}
				}
				
				receiveStatus = STATUS_IDLE;
				
				// Wait for NACK, clear interrupt flag
				TWCR |= (1<<TWINT)|(0<<TWEA);	
				
			}else {
				receiveStatus = STATUS_IDLE;
				// Wait for NACK, clear interrupt flag
				TWCR |= (1<<TWINT)|(0<<TWEA);	
			}
		}break;
		
		// Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
		case 0x88 : {
			if (receiveStatus == STATUS_POINTER) {
				// First byte of transmission, register pointer
				registerPointer = TWDR;
				if (registerPointer >= SERVO_DATA_SIZE) {
					registerPointer = 0;
				}
				receiveStatus = STATUS_DATA;
				
			}else if (receiveStatus == STATUS_DATA) {
				// Second byte of transmission, servo data
				servoData[registerPointer] = TWDR;
				updateFlag = 1;
				receiveStatus = STATUS_IDLE;
				
			}else {
				receiveStatus = STATUS_IDLE;
			}
			
			// Wait for ACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);	
		}break;
		
		// A STOP condition or repeated START condition has been received while still addressed as Slave
		case 0xA0 : {
			// Master wants to read a byte, which should now be pointed to by the registerPointer
			// Wait for ACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		
		
		/*************************************** Slave Transmitt mode***********************************/
		
		// Own SLA+R has been received; ACK has been returned
		case 0xA8 : {
			// Send data byte
			TWDR = servoData[registerPointer];
			// Wait for NACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(0<<TWEA);
		}break;
		
		// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
		case 0xB0 : {
			// Send first data byte
			TWDR = servoData[0];
			// Wait for NACK, clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		// Data byte in TWDR has been transmitted; ACK has been received
		case 0xB8 : {
			// Whatevs Master, we're done, so return to addressable slave I2C mode
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		// Data byte in TWDR has been transmitted; NOT ACK has been received
		case 0xC0 : {
			// Good, we're done, so return to addressable slave I2C mode
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		// Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received
		case 0xC8 : {
			// Whatevs Master, we're done, so return to addressable slave I2C mode
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}break;
		
		/*************************************************************************************************/
		
		default : {
			// Clear interrupt flag
			TWCR |= (1<<TWINT)|(1<<TWEA);
		}
	}
}


// Fills the correct time into the servo output array
void calculateServoTimes(void) {
	uint8_t i, j;
	// Calculate the servoTimes
	i = 0;
	while(i!=255) {
		for(j=0; j<=SERVO_ADR_END; j++) {
		
			// Block servo 0 if disabled
			if (!servoData[SERVO0_EN] && j==SERVO_0_ADR) {
				// Fill all servo 0 times with 0
				servoTimes[i] &= ~(1<<j);
				
			}else {
				// Fill times with data
				if (servoData[j] > i) {
					servoTimes[i] |= (1<<j);
				}else {
					servoTimes[i] &= ~(1<<j);
				}
			}
			
		}
		
		i++;
		
		/*if (i!=255) {
			i++;
		}else {
			break;
		}*/
	}
}

// Reads the temperature from the DS18B20
uint16_t therm_read_temperature(void)
{
	uint16_t temp;

	cli();
	//Reset, skip ROM and start temperature conversion
	therm_reset();
	therm_write_byte(THERM_CMD_SKIPROM);
	therm_write_byte(THERM_CMD_CONVERTTEMP);
	sei();

	//Wait until conversion is complete
	while(!therm_read_bit());
	
	cli();
	//Reset, skip ROM and send command to read Scratchpad
	therm_reset();
	therm_write_byte(THERM_CMD_SKIPROM);
	therm_write_byte(THERM_CMD_RSCRATCHPAD);
	
	//Read Scratchpad (only 2 first bytes)
	temp = therm_read_byte();
	temp |= (therm_read_byte() << 8);
	
	therm_reset();
	sei();
	
	return temp;
}

// Main Program Loop
int main(void)
{
	uint8_t time = 0;
	uint8_t buffered_temp_flag = 0;
	uint16_t temperature_reading;

	// PORTB Initialization, status LED
	PORTB = 0xFF;
	DDRB = (1<<PB1);

	// PORTC Initialization, set as input
	PORTC = 0xFF;
	//DDRC = 0;
	// Set I2C pins as outputs
	DDRC = (1<<PC4)|(1<<PC5);

	// PORTD Initialization, servo outputs
	PORTD = 0xFF;
	DDRD = (1<<PD2)|(1<<PD1)|(1<<PD0);
	
	// Initialize I2C, slave transmitt mode only
	// Set slave address
	TWAR = I2C_SLAVE_ADDRESS;
	// Set control register, ack on slave address, and enable TWI. Also interrupts.
	TWCR = (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|(1<<TWEN)|(1<<TWIE);
	
	// The register pointer used to set where to read and write from.
	registerPointer = 0;
	receiveStatus = STATUS_IDLE;
	updateFlag = 0;
	updateTemperatureFlag = 0;
	buffered_temp_flag = 0;
	temperature_reading = 0;
	
	// Initialize timer 1 for 20ms interrupts, used to set timing for servo update signals
	TCCR1B = (1<<WGM12)|(0<<CS12)|(1<<CS11)|(0<<CS10);	// CTC mode, 1:8 prescaler so timer clock of 2MHz
	OCR1A = 40000;			// 20ms interval
	TIMSK1 = (1<<OCIE1A);
	
	// Initialize servos to midpoint
	for(time=0; time<=SERVO_ADR_END; time++) {
		servoData[time] = 127;
	}
	calculateServoTimes();
	
	// Disable servo 0 by default
	servoData[SERVO0_EN] = 0;
	
	led_counter = 0;
	
	// Init DS18B20 temperature sensor
	// High temp: 60C, low temp: -30C, Resolution: 11-bits for 375ms processing time
	therm_config(60, (128 | 30), 0b01011111);
	
	// Enable global interrupts
	sei();

	while(1)
	{
		// Read pin state for IR beam sensor
		servoData[IR_BEAM] = (~PIND & IR_SENS_PIN);
		
		// Waveform generation
		loopflag = 1;
		buffered_temp_flag = updateTemperatureFlag;
		
		// Start creating the servo timing signal
		// Block servo 0 if disabled
		if (servoData[SERVO0_EN]) {
			PORTD = (1<<PD2)|(1<<PD1)|(1<<PD0);
		}else {
			PORTD = (1<<PD2)|(1<<PD1);
		}
		
		_delay_us(1000);
		
		// Let it do it's work for the next 1000us
		time = 0;
		do {
			time++;
			// 1ms / 255 = 3,92s
			// 16.00MHz * 3,92s = 62.75
			// 62.75 clock cycles are 3,93s, so waste 63 clock cycles before the next loop
			// However, there is a bunch of overhead in the code, so this actually needs to be tuned
			asm volatile("LDI R18, 20");
			asm volatile("LOOP: DEC R18");
			asm volatile("BRNE LOOP");
			PORTD = servoTimes[time];
			
		} while(time != 255);
		PORTD = 0;
		
		// This flag is cleared every 20ms
		while(loopflag) {
			if (updateFlag) {
				calculateServoTimes();
				updateFlag = 0;
			}
			
			// Get new temperature reading from 
			if (buffered_temp_flag) {
				// Get temperature sensor reading
				// Read temp, store in global variable for later
				temperature_reading = therm_read_temperature();
				
				servoData[TEMP_SENS] = temperature_reading;
				servoData[TEMP_SENS+1] = (temperature_reading >> 8);
				
				// Clear flags
				buffered_temp_flag = 0;
				updateTemperatureFlag = 0;
			}
		};
		
	};
	return 0;
}
